/*
 * File: display.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */


/***

The functions in this file handle redisplay. There are two pieces, the
ones that update the virtual display screen, and the ones that make the
physical display screen the same as the virtual display screen. These
functions use hints that are left in the windows by the commands.

***/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"rho.h"
#include	"display.h"
#include	"cmode.h"
#include	"line.h"
#include	"messages.h"
#include	"terminal.h"
#include	"utils.h"



static VIDEO **vscreen;         /* Virtual screen. */
static VIDEO **pscreen;         /* Physical screen. */


/*++

Initialize the data structures used by the display code. The edge vectors
used to access the screens are set up. The operating system's terminal
I/O channel is set up. Everything else gets initialized at compile time.
The original window has "WFCHG" set, so that it will get completely
redrawn on the first call to "update".

++*/

int vtinit()
{
    int i;
    VIDEO *vp;

    TTopen();                   /* open the screen */
    TTkopen();                  /* open the keyboard */
    TTrev(FALSE);

    /* allocate the virtual screen pointer array */
    vscreen = (VIDEO **) malloc(term.t_mrow * sizeof(VIDEO *));

    /* allocate the physical shadow screen array */
    pscreen = (VIDEO **) malloc(term.t_mrow * sizeof(VIDEO *));

    /* for every line in the display */
    for (i = 0; i < term.t_mrow; ++i) {

        /* allocate a virtual screen line */
        vp = (VIDEO *) malloc(sizeof(VIDEO) + term.t_mcol);

        vp->v_flag = 0;         /* init change flags */
        vp->v_left = FARRIGHT;  /* init requested rev video boundries */
        vp->v_right = 0;

        /* connect virtual line to line array */
        vscreen[i] = vp;

        /* allocate and initialize physical shadow screen line */
        vp = (VIDEO *) malloc(sizeof(VIDEO) + term.t_mcol);

        vp->v_flag = VFNEW;
        vp->v_left = FARRIGHT;
        vp->v_right = 0;
        pscreen[i] = vp;
    }
    return (TRUE);
}



/*++

Send a command to the terminal to move the hardware cursor to row
"row" and column "col". The row and column arguments are origin 0,0.
Only real moves. Update "ttrow" and "ttcol".

++*/
void movecursor(row, col)
int row, col;

{
    /* only move it if there is a difference */
    if (row != ttrow || col != ttcol) {
        ttrow = row;
        ttcol = col;
        TTmove(row, col);
    }
}



/*++

Clean up the virtual terminal system in anticipation for a return to
the operating system. Move down to the last line and clear it out (the
next system prompt will be written on that line). Shut down the channel
to the terminal.

++*/
void vttidy()
{
    mlerase();
    movecursor(term.t_nrow, 0);
    TTflush();
    TTclose();
    TTkclose();
}



/*++

Set the virtual cursor to the specified row and column on the virtual
screen.

++*/
static void vtmove(row, col)
int row, col;

{
    vtrow = row;
    vtcol = col;
}



/*++

Write a character to the virtual screen. The virtual row and column
are updated. If we are not yet on left edge, don't print it yet. If
the line is too long put a "$" in the last column.  This routine only
puts printing characters into the virtual terminal buffers. Only column
overflow is checked.

++*/
static void vtputc(c)
int c;

{
    VIDEO *vp;                  /* ptr to line being updated */

    /* defeate automatic sign extenstion */
    c = c & 0xff;

    /* this is the line to put this character! */
    vp = vscreen[vtrow];

    if (c == '\t') {

        /* output a hardtab as the right number of spaces */
        do {
            vtputc(' ');
        }
        while (((vtcol + taboff) % (hardtab)) != 0);

    } else if (vtcol >= term.t_ncol) {

        /* we are at the right edge! */
        ++vtcol;
        vp->v_text[term.t_ncol - 1] = '$';

    } else if (c < 0x20 || c == 0x7F) {

        /* control character? */
        vtputc('^');
        vtputc(c ^ 0x40);

    } else {

        /* it's normal, just put it in the screen map */
        if (vtcol >= 0)
            vp->v_text[vtcol] = c;
        ++vtcol;
    }
}



/*++

Erase from the end of the software cursor to end of the line.
Local to display.c

++*/
static void vteeol()
{
    VIDEO *vp;

    vp = vscreen[vtrow];
    while (vtcol < term.t_ncol) {
        if (vtcol >= 0)
            vp->v_text[vtcol] = ' ';
        vtcol++;
    }
}



/*++

Check to see if the cursor is on in the window and re-frame it if needed
or wanted. Helper for update()

++*/
static void reframe(wp)
WNDW *wp;

{
    LINE *lp;                   /* search pointer */
    LINE *rp;                   /* reverse search pointer */
    LINE *hp;                   /* ptr to header line in buffer */
    LINE *tp;                   /* temp debugging pointer */
    int i;                      /* general index/# lines to scroll */
    int nlines;                 /* number of lines in current window */

    /* figure out our window size */
    nlines = wp->w_ntrows;
    if (modeflag == FALSE)
        nlines++;

    /* if not a requested reframe, check for a needed one */
    if ((wp->w_flag & WFFORCE) == 0) {
        lp = wp->w_linep;
        for (i = 0; i < nlines; i++) {

            /* if the line is in the window, no reframe */
            if (lp == wp->w_dotp)
                return;

            /* if we are at the end of the file, reframe */
            if (lp == wp->w_bufp->b_linep)
                break;

            /* on to the next line */
            lp = lforw(lp);
        }
    }

    /* reaching here, we need a window refresh */
    i = wp->w_force;

    /* if smooth scrolling is enabled,
       first.. have we gone off the top? */
    if (sscroll && ((wp->w_flag & WFFORCE) == 0)) {
        /* search thru the buffer looking for the point */
        tp = lp = rp = wp->w_linep;
        hp = wp->w_bufp->b_linep;
        while ((lp != hp) || (rp != hp)) {

            /* did we scroll downward? */
            if (lp == wp->w_dotp) {
                i = nlines - 1;
                break;
            }

            /* did we scroll upward? */
            if (rp == wp->w_dotp) {
                i = 0;
                break;
            }

            /* advance forward and back */
            if (lp != hp)
                lp = lforw(lp);
            if (rp != hp)
                rp = lback(rp);

            /* problems????? */
            if (lp == tp || rp == tp) {
                mlforce("BUG IN SMOOTH SCROLL--GET DAN!\n");
                TTgetc();
            }
        }
        /* how far back to reframe? */
    } else if (i > 0) {         /* only one screen worth of lines max */
        if (--i >= nlines)
            i = nlines - 1;
    } else if (i < 0) {         /* negative update???? */
        i += nlines;
        if (i < 0)
            i = 0;
    } else
        i = nlines / 2;

    /* backup to new line at top of window */
    lp = wp->w_dotp;
    while (i != 0 && lback(lp) != wp->w_bufp->b_linep) {
        --i;
        if (i < 0) {
            mlforce("OTHER BUG IN DISPLAY --- GET DAN!!!\n");
            TTgetc();
        }
        lp = lback(lp);
    }

    /* and reset the current line at top of window */
    wp->w_linep = lp;
    wp->w_flag |= WFHARD;
    wp->w_flag &= ~WFFORCE;
}



/*++

Update the current line to the virtual screen.
Helper for update()

++*/
static void updcurln(wp)
WNDW *wp;                       /* window to update current line in */

{
    LINE *lp;                   /* line to update */
    int sline;                  /* physical screen line to update */
    int i;

    /* search down the line we want */
    lp = wp->w_linep;
    sline = wp->w_toprow;
    while (lp != wp->w_dotp) {
        ++sline;
        lp = lforw(lp);
    }

    /* and update the virtual line */
    vscreen[sline]->v_flag |= VFCHG;
    taboff = wp->w_fcol;
    vtmove(sline, -taboff);

    /* move each char of line to virtual screen until at end */
    for (i = 0; i < lused(lp); ++i)
        vtputc(lgetc(lp, i));
    vteeol();
    taboff = 0;
}




/*++

Update all lines in a window on the virtual screen.
Helper for update()

++*/
static void updall(wp)
WNDW *wp;                       /* window to update lines in */

{
    LINE *lp;                   /* line to update */
    int sline;                  /* physical screen line to update */
    int i;
    int nlines;                 /* number of lines in the current window */

    /* search down the lines, updating them */
    lp = wp->w_linep;
    sline = wp->w_toprow;
    nlines = wp->w_ntrows;
    if (modeflag == FALSE)
        nlines++;
    taboff = wp->w_fcol;
    while (sline < wp->w_toprow + nlines) {

        /* and update the virtual line */
        vscreen[sline]->v_flag |= VFCHG;
        vscreen[sline]->v_left = FARRIGHT;
        vscreen[sline]->v_right = 0;
        vtmove(sline, -taboff);
        if (lp != wp->w_bufp->b_linep) {
            /* if we are not at the end */
            for (i = 0; i < lused(lp); ++i)
                vtputc(lgetc(lp, i));
            lp = lforw(lp);
        }

        /* make sure we are on screen */
        if (vtcol < 0)
            vtcol = 0;

        /* on to the next one */
        vteeol();
        ++sline;
    }
    taboff = 0;
}



/*++

Update the current extended line at a column greater than the terminal
width. The line will be scrolled right or left to let the user see where
the cursor is.

++*/
static void updext()
{
    int rcursor;                /* real cursor location */
    LINE *lp;                   /* pointer to current line */
    int j;                      /* index into line */

    /* calculate what column the real cursor will end up in */
    rcursor = ((curcol - term.t_ncol) % term.t_scrsiz) + term.t_margin;
    lbound = curcol - rcursor + 1;
    taboff = lbound + curwp->w_fcol;

    /* scan through the line outputing characters to the virtual screen */
    /* once we reach the left edge                                  */
    vtmove(currow, -taboff);    /* start scanning offscreen */
    lp = curwp->w_dotp;         /* line to output */
    for (j = 0; j < lused(lp); ++j)     /* until the end-of-line */
        vtputc(lgetc(lp, j));

    /* truncate the virtual line, restore tab offset */
    vteeol();
    taboff = 0;

    /* and put a '$' in column 1 */
    vscreen[currow]->v_text[0] = '$';
}



/*++

Update the position of the hardware cursor and handle extended lines. This
is the only update for simple moves. Helper for update()

++*/
static void updpos()
{
    LINE *lp;
    int c;
    int i;

    /* find the current row */
    lp = curwp->w_linep;
    currow = curwp->w_toprow;
    while (lp != curwp->w_dotp) {
        ++currow;
        lp = lforw(lp);
    }

    /* find the current column */
    curcol = 0;
    i = 0;
    while (i < curwp->w_doto) {
        c = lgetc(lp, i++);
        if (c == '\t')
            curcol += -(curcol % hardtab) + (hardtab - 1);
        else {
            if (c < 0x20 || c == 0x7f)
                ++curcol;
        }
        if (c == ectoc(c))
            ++curcol;
    }

    /* adjust by the current first column position */
    curcol -= curwp->w_fcol;

    /* make sure it is not off the left side of the screen */
    while (curcol < 0) {
        if (curwp->w_fcol >= hjump) {
            curcol += hjump;
            curwp->w_fcol -= hjump;
        } else {
            curcol += curwp->w_fcol;
            curwp->w_fcol = 0;
        }
        curwp->w_flag |= WFHARD | WFMODE;
    }

    /* if horizontall scrolling is enabled, shift if needed */
    if (hscroll) {
        while (curcol >= term.t_ncol - 1) {
            curcol -= hjump;
            curwp->w_fcol += hjump;
            curwp->w_flag |= WFHARD | WFMODE;
        }
    } else {
        /* if extended, flag so and update the virtual line image */
        if (curcol >= term.t_ncol - 1) {
            vscreen[currow]->v_flag |= (VFEXT | VFCHG);
            updext();
        } else
            lbound = 0;
    }
}



/*++

De-extend any line that needs it.

++*/
static void upddex()
{
    WNDW *wp;
    LINE *lp;
    int i, j;
    int nlines;                 /* number of lines in the current window */

    wp = wheadp;

    while (wp != NULL) {
        lp = wp->w_linep;
        i = wp->w_toprow;
        nlines = wp->w_ntrows;
        if (modeflag == FALSE)
            nlines++;

        while (i < wp->w_toprow + nlines) {
            if (vscreen[i]->v_flag & VFEXT) {
                if ((wp != curwp) || (lp != wp->w_dotp) ||
                    (curcol < term.t_ncol - 1)) {
                    taboff = wp->w_fcol;
                    vtmove(i, -taboff);
                    for (j = 0; j < lused(lp); ++j)
                        vtputc(lgetc(lp, j));
                    vteeol();
                    taboff = 0;

                    /* this line no longer is extended */
                    vscreen[i]->v_flag &= ~VFEXT;
                    vscreen[i]->v_flag |= VFCHG;
                }
            }
            lp = lforw(lp);
            ++i;
        }
        /* and onward to the next window */
        wp = wp->w_wndp;
    }
}




/*++

If the screen is garbage, clear the physical screen and the virtual
screen and force a full update.

++*/
static void updgar()
{
    int i;
    int j;
    char *txt;

    for (i = 0; i < term.t_nrow; ++i) {
        vscreen[i]->v_flag |= VFCHG;
        pscreen[i]->v_left = FARRIGHT;
        pscreen[i]->v_right = 0;
        txt = pscreen[i]->v_text;
        for (j = 0; j < term.t_ncol; ++j)
            txt[j] = ' ';
        pscreen[i]->v_flag &= ~VFNEW;
    }

    movecursor(0, 0);           /* Erase the screen. */
    TTeeop();
    sgarbf = FALSE;             /* Erase-page clears */
    mpresf = FALSE;             /* the message area. */
}




/*++

Update a single line. This does not know how to use insert- or delete-
character sequences; we are using VT52 functionality. Update the physical
row and column variables. It does try to erase to end of line.

++*/
static void update_line(row, vp, pp)
int row;                        /* row of screen to update */
struct VIDEO *vp;               /* virtual screen image */
struct VIDEO *pp;               /* physical screen image */

{

    char *vir_left;             /* left pointer to virtual line */
    char *phy_left;             /* left pointer to physical line */
    char *vir_right;            /* right pointer to virtual line */
    char *phy_right;            /* right pointer to physical line */
    int rev_left;               /* leftmost reversed char index */
    int rev_right;              /* rightmost reversed char index */
    char *left_blank;           /* left-most trailing blank */
    int non_blanks;             /* non-blanks to the right flag */
    int update_column;          /* update column */
    int old_rev_state = FALSE;  /* reverse video states */
    int new_rev_state;

    /* set up pointers to virtual and physical lines */
    vir_left = &vp->v_text[0];
    vir_right = &vp->v_text[term.t_ncol];
    phy_left = &pp->v_text[0];
    phy_right = &pp->v_text[term.t_ncol];
    update_column = 0;
    rev_left = FARRIGHT;
    rev_right = 0;
    non_blanks = TRUE;

    /* if this is a legitimate line to optimize */
    if (!(pp->v_flag & VFNEW)) {

        /* advance past any common chars at the left */
        while ((vir_left != &vp->v_text[term.t_ncol])
               && (vir_left[0] == phy_left[0])) {
            ++vir_left;
            ++update_column;
            ++phy_left;
        }

        /* advance past any common chars at the right */
        non_blanks = FALSE;
        while ((vir_right[-1] == phy_right[-1]) && (vir_right >= vir_left)) {
            --vir_right;
            --phy_right;

            /* Note if any nonblank in right match. */
            if (vir_right[0] != ' ')
                non_blanks = TRUE;
        }

    }


    /* reverse video changes? */
    if ((vp->v_left != pp->v_left) || (vp->v_right != pp->v_right)) {

        /* adjust leftmost edge */
        if (vp->v_left < pp->v_left)
            rev_left = vp->v_left;
        else
            rev_left = pp->v_left;
        pp->v_left = vp->v_left;
        if (rev_left < update_column) {
            vir_left = &vp->v_text[rev_left];
            phy_left = &pp->v_text[rev_left];
            update_column = rev_left;
        }

        /* adjust rightmost edge */
        if (vp->v_right > pp->v_right)
            rev_right = vp->v_right;
        else
            rev_right = pp->v_right;
        pp->v_right = vp->v_right;
        if (&vp->v_text[rev_right] > vir_right) {
            vir_right = &vp->v_text[rev_right];
            phy_right = &pp->v_text[rev_right];
        }
    } else {
        rev_left = vp->v_left;
        rev_right = vp->v_right;
    }

    /* if both lines are the same, no update needs to be done */
    if (!(pp->v_flag & VFNEW) && (vir_left > vir_right)) {
        vp->v_flag &= ~VFCHG;   /* flag this line as changed */
        return;
    }

    left_blank = vir_right;

    /* Erase to EOL ? */
    if (non_blanks == FALSE && eolexist == TRUE) {
        while (left_blank != vir_left && left_blank[-1] == ' ')
            --left_blank;

        if (vir_right - left_blank <= 3)        /* Use only if erase is */
            left_blank = vir_right;     /* fewer characters. */
    }

    /* move to the beginning of the text to update */
    movecursor(row, update_column);

    while (vir_left != left_blank) {    /* Ordinary. */

        /* are we in a reverse video field? */
        if (pp->v_left <= update_column && update_column < pp->v_right)
            new_rev_state = TRUE;
        else
            new_rev_state = FALSE;

        /* if moving in or out of rev video, change it */
        if (new_rev_state != old_rev_state) {
            TTrev(new_rev_state);
            old_rev_state = new_rev_state;
        }

        /* output the next character! */
        TTputc(*vir_left);
        ++ttcol;
        ++update_column;
        *phy_left++ = *vir_left++;
    }

    if (left_blank != vir_right) {      /* Erase. */

        /* are we in a reverse video field? */
        if (pp->v_left <= update_column && update_column < pp->v_right)
            new_rev_state = TRUE;
        else
            new_rev_state = FALSE;

        /* if moving in or out of rev video, change it */
        if (new_rev_state != old_rev_state) {
            TTrev(new_rev_state);
            old_rev_state = new_rev_state;
        }

        /*
         * TERMCAP does not tell us if the current terminal propagates
         * the current attributes to the end of the line when an erase
         * to end of line sequence is sent. Don't let TERMCAP use EEOL
         * if in a reverse video line.
         */
        if (new_rev_state == TRUE)
            while (update_column++ < term.t_ncol)
                TTputc(' ');
        else
            TTeeol();
        while (vir_left != vir_right)
            *phy_left++ = *vir_left++;
    }

    vp->v_flag &= ~VFCHG;       /* flag this line as updated */
    vp->v_flag &= ~VFCOL;

    /* Always leave in the default state */
    if (old_rev_state == TRUE)
        TTrev(FALSE);
    return;
}



/*++

Update the physical screen from the virtual screen.
Global helper.

++*/
static void updupd(force)
int force;                      /* forced update flag */

{
    VIDEO *vp1;
    int i;

    for (i = 0; i < term.t_nrow; ++i) {
        vp1 = vscreen[i];

        /* for each line that needs to be updated */
        if (vp1->v_flag & VFCHG) {
            if (force == FALSE && type_ahead())
                return;
            update_line(i, vp1, pscreen[i]);
        }
    }
    return;
}



/*++

Redisplay the mode line for the window pointed to by the "wp". This is
the only routine that has any idea of how the modeline is formatted.
Called by "update" any time there is a dirty window.

++*/
static void modeline(wp)
WNDW *wp;                       /* window to update modeline for */

{
    BUFFER *bp;
    char *cp;
    char c;
    char tline[NLINE];          /* buffer for part of mode line */
    int firstm;                 /* is this the first mode? */
    int i;                      /* loop index */
    int lchar;                  /* character to draw line in buffer with */
    int n;                      /* cursor position count */

    if (modeflag == FALSE)
        return;

    n = wp->w_toprow + wp->w_ntrows;    /* Location. */

/*
	Note that we assume that setting REVERSE will cause the terminal
	driver to draw with the inverted relationship of fcolor and
	bcolor, so that when we say to set the foreground color to "white"
	and background color to "black", the fact that "reverse" is
	enabled means that the terminal driver actually draws "black" on a
	background of "white".  This way, devices for which the color controls
	are optional will still get the "reverse" signals.
*/

    vscreen[n]->v_flag |= VFCHG | VFCOL;        /* Redraw next time. */
    vscreen[n]->v_left = 0;
    vscreen[n]->v_right = term.t_ncol;
    vtmove(n, 0);               /* Seek to right line. */
    if (wp == curwp)            /* mark the current buffer */
        lchar = '=';
    else if (revexist)
        lchar = ' ';
    else
        lchar = '-';

    bp = wp->w_bufp;
    if ((bp->b_flag & BFTRUNC) != 0)    /* "#" if truncated */
        vtputc('#');
    else
        vtputc(lchar);

    if ((bp->b_flag & BFCHG) != 0)      /* "*" if changed. */
        vtputc('*');
    else
        vtputc(lchar);

    if ((bp->b_flag & BFNAROW) != 0) {  /* "<>" if narrowed */
        vtputc('<');
        vtputc('>');
    } else {
        vtputc(lchar);
        vtputc(lchar);
    }

    n = 4;
    strcpy(tline, " ");         /* Buffer name. */
    strcat(tline, PROGNAME);
    strcat(tline, " ");
    strcat(tline, VERSION);
    strcat(tline, " ");
    strcat(tline, "== ");

    /* are we horizontally scrolled? */
    if (wp->w_fcol > 0) {
        strcat(tline, "[<");
        strcat(tline, int2str(wp->w_fcol));
        strcat(tline, "]");
    }

    /* display the point position in buffer if on current modeline */
    if (posflag && wp == curwp) {

        strcat(tline, "L:");
        strcat(tline, int2str(getlinenum(bp, wp->w_dotp)));
        strcat(tline, " C:");
        strcat(tline, int2str(1 + getccol(FALSE)));
        strcat(tline, " ");
    }

    /* display the modes */
    strcat(tline, "== ");
    firstm = TRUE;
    for (i = 0; i < NUMMODES; i++)      /* add in the mode flags */
        if (wp->w_bufp->b_mode & (1 << i)) {
            if (firstm != TRUE)
                strcat(tline, ":");
            firstm = FALSE;
            strcat(tline, modename[i]);
        }
    strcat(tline, " ");

    cp = &tline[0];
    while ((c = *cp++) != 0) {
        vtputc(c);
        ++n;
    }

    vtputc(lchar);
    vtputc(lchar);
    n += 3;

    if (bp->b_name[0] != 0) {   /* File name. */
        vtputc(' ');
        ++n;

        cp = &bp->b_name[0];

        while ((c = *cp++) != 0) {
            vtputc(c);
            ++n;
        }

        vtputc(' ');
        ++n;
    }

    while (n < term.t_ncol) {   /* Pad to full width. */
        vtputc(lchar);
        ++n;
    }
}



/*++

Make sure that the display is right. This is a three part process. First,
scan through all of the windows looking for dirty ones. Check the framing,
and refresh the screen. Second, make sure that "currow" and "curcol" are
correct for the current window. Third, make the virtual and physical
screens the same.

++*/
void update(force)
int force;                      /* force update past type-ahead? */

{
    WNDW *wp;

    /* if we are not forcing the update, and there are keystrokes
       waiting to be processed, skip the update */
    if (force == FALSE && type_ahead())
        return;

    /* if we are replaying a keyboard macro, don't bother keeping updated */
    if (force == FALSE && kbdmode == PLAY)
        return;

    /* update any windows that need refreshing */
    wp = wheadp;
    while (wp != NULL) {
        if (wp->w_flag) {
            /* if the window has changed, service it */
            reframe(wp);        /* check the framing */
            if ((wp->w_flag & ~WFMODE) == WFEDIT)
                updcurln(wp);   /* update EDITed line */
            else if (wp->w_flag & ~WFMOVE)
                updall(wp);     /* update all lines */
            if (wp->w_flag & WFMODE)
                modeline(wp);   /* update modeline */
            wp->w_flag = 0;
            wp->w_force = 0;
        }

        /* on to the next window */
        wp = wp->w_wndp;
    }

    /* recalc current hardware cursor location */
    updpos();

    /* update the current window if we have to move it around */
    if (curwp->w_flag & WFHARD)
        updall(curwp);
    if (curwp->w_flag & WFMODE)
        modeline(curwp);
    curwp->w_flag = 0;

    /* check for lines to de-extend */
    upddex();

    /* if screen is garbage, re-plot it */
    if (sgarbf != FALSE) {
        if (gflags & GFSDRAW) {
            sgarbf = FALSE;
        } else {
            updgar();
        }
    }

    /* update the virtual screen to the physical screen */
    updupd(force);

    /* update the cursor and flush the buffers */
    movecursor(currow, curcol - lbound);
    TTflush();

    return;
}



/*++

Update all the modelines.

++*/
void update_modeline()
{
    WNDW *wp;
    wp = wheadp;
    while (wp != NULL) {
        wp->w_flag |= WFMODE;
        wp = wp->w_wndp;
    }
}



/*++

Force hard updates on all windows.

++*/
void update_window()
{
    WNDW *wp;
    wp = wheadp;
    while (wp != NULL) {
        wp->w_flag |= WFHARD | WFMODE;
        wp = wp->w_wndp;
    }
}



/*++

This function initializes our only screen.  The text is in a buffer.
The one screen is the size of the terminal you are using.  So you can
see that much of the buffer. And windows are a potentially restricted
view onto that screen.  Multiple windows on the single screen view two
or more window-sized chunks of one or more buffers.

++*/
SCRN *init_screen(scr_name, scr_buf)
char *scr_name;                 /* screen name */
BUFFER *scr_buf;                /* buffer to place in first window of screen */

{
    int cmark;                  /* current mark to initialize */
    SCRN *sp;                   /* pointer to allocated screen */
    SCRN *last_sp;              /* pointer to last screen */
    WNDW *wp;                   /* ptr to first window of new screen */

    /* allocate memory for this screen */
    sp = (SCRN *) malloc(sizeof(SCRN));



    /* set up this new screens fields! */
    sp->s_next_screen = (SCRN *) NULL;
    sp->s_screen_name = strdup(scr_name);
    sp->s_roworg = term.t_roworg;
    sp->s_colorg = term.t_colorg;
    sp->s_nrow = term.t_nrow;
    sp->s_ncol = term.t_ncol;

    /* allocate its first window */
    wp = (WNDW *) malloc(sizeof(WNDW));

    sp->s_first_window = sp->s_cur_window = wp;

    /* and setup the windows info */
    wp->w_wndp = NULL;
    wp->w_bufp = scr_buf;
    scr_buf->b_nwnd += 1;
    wp->w_linep = scr_buf->b_linep;

    /* position us at the buffers dot */
    wp->w_dotp = scr_buf->b_dotp;
    wp->w_doto = scr_buf->b_doto;

    /* set all the marks to UNSET */
    for (cmark = 0; cmark < NMARKS; cmark++) {
        wp->w_markp[cmark] = NULL;
        wp->w_marko[cmark] = 0;
    }
    wp->w_toprow = 0;
    wp->w_fcol = 0;
    wp->w_ntrows = term.t_nrow - 1;     /* "-1" for mode line.  */
    wp->w_force = 0;
    wp->w_flag = WFMODE | WFHARD;       /* Full. */

    /* first screen? */
    if (first_screen == (SCRN *) NULL) {
        first_screen = sp;
        return (sp);
    }

    /* insert it at the tail of the screen list */
    last_sp = first_screen;
    while (last_sp->s_next_screen != (SCRN *) NULL)
        last_sp = last_sp->s_next_screen;
    last_sp->s_next_screen = sp;

    /* and return the new screen pointer */
    return (sp);
}



/*++

User routine to force a screen update.  Always finishes complete update.

++*/
int update_screen(f, n)
int f, n;

{
    f = n = 0;                  /* splint */

    update(TRUE);
    return (TRUE);
}
